@@ -1,5 +1,5 @@
 # -*- coding: utf-8 -*-
-# Copyright (c) 2003-2009 LOGILAB S.A. (Paris, FRANCE).
+# Copyright (c) 2003-2010 LOGILAB S.A. (Paris, FRANCE).
 # http://www.logilab.fr/ -- mailto:contact@logilab.fr
 #
 # This program is free software; you can redistribute it and/or modify it under
@@ -226,6 +226,20 @@ class HgConfig(object):
         """
         return self.ui.config(self.section, 'diffsectioncolor', default)
 
+    @cached
+    def getMQFGColor(self, default='#ff8183'):
+        """
+        mqfgcolor: bg color to highlight mq patches
+        """
+        return self.ui.config(self.section, 'mqfgcolor', default)
+
+    @cached
+    def getMQHideTags(self, default=False):
+        """
+        mqhidetags: hide mq tags
+        """
+        return self.ui.config(self.section, 'mqhidetags', default)
+
 
 
 _HgConfig = HgConfig
