@@ -1,4 +1,4 @@
-# Copyright (c) 2003-2009 LOGILAB S.A. (Paris, FRANCE).
+# Copyright (c) 2003-2010 LOGILAB S.A. (Paris, FRANCE).
 # http://www.logilab.fr/ -- mailto:contact@logilab.fr
 #
 # This program is free software; you can redistribute it and/or modify it under
@@ -24,10 +24,10 @@ from cStringIO import StringIO
 import difflib
 
 from mercurial.node import nullrev
-from mercurial import patch, util
+from mercurial import patch, util, match
 
 import hgviewlib # force apply monkeypatches
-from hgviewlib.util import tounicode
+from hgviewlib.util import tounicode, isbfile
 
 def diff(repo, ctx1, ctx2=None, files=None):
     """
@@ -36,19 +36,18 @@ def diff(repo, ctx1, ctx2=None, files=None):
     if ctx2 is None:
         ctx2 = ctx1.parents()[0]
     if files is None:
-        match = util.always
+        m = match.always(repo.root, repo.getcwd())
     else:
-        def match(fname):
-            return fname in files
+        m = match.exact(repo.root, repo.getcwd(), files)
     # try/except for the sake of hg compatibility (API changes between
     # 1.0 and 1.1)
     try:
         out = StringIO()
-        patch.diff(repo, ctx2.node(), ctx1.node(), match=match, fp=out)
+        patch.diff(repo, ctx2.node(), ctx1.node(), match=m, fp=out)
         diffdata = out.getvalue()
     except:
         diffdata = '\n'.join(patch.diff(repo, ctx2.node(), ctx1.node(),
-                                        match=match))
+                                        match=m))
     # XXX how to deal diff encodings?
     try:
         diffdata = unicode(diffdata, "utf-8")
@@ -65,7 +64,12 @@ def __get_parents(repo, rev, branch=None):
     much slower).
     """
     if not branch:
+        if rev is None:
+            return [x.rev() for x in repo.changectx(None).parents() if x]
         return [x for x in repo.changelog.parentrevs(rev) if x != nullrev]
+    if rev is None:
+        return [x.rev() for x in repo.changectx(None).parents() \
+                if x and repo.changectx(rev).branch() == branch]
     return [x for x in repo.changelog.parentrevs(rev) \
             if (x != nullrev and repo.changectx(rev).branch() == branch)]
 
@@ -79,7 +83,7 @@ def ismerge(ctx):
         return len(ctx.parents()) == 2 and ctx.parents()[1]
     return False
 
-def revision_grapher(repo, start_rev=None, stop_rev=0, branch=None):
+def revision_grapher(repo, start_rev=None, stop_rev=0, branch=None, follow=False):
     """incremental revision grapher
 
     This generator function walks through the revision history from
@@ -87,31 +91,41 @@ def revision_grapher(repo, start_rev=None, stop_rev=0, branch=None):
     or equal to start_rev) and for each revision emits tuples with the
     following elements:
 
-      - Current revision.
-      - Column of the current node in the set of ongoing edges.
+      - current revision
+      - column of the current node in the set of ongoing edges
       - color of the node (?)
       - lines; a list of (col, next_col, color) indicating the edges between
         the current row and the next row
       - parent revisions of current revision
 
+    If follow is True, only generated the subtree from the start_rev head.
+
+    If branch is set, only generated the subtree for the given named branch. 
     """
-    if start_rev is None:
+    if start_rev is None and repo.status() == ([],)*7:
         start_rev = len(repo.changelog)
-    assert start_rev >= stop_rev
+    assert start_rev is None or start_rev >= stop_rev
     curr_rev = start_rev
     revs = []
     rev_color = {}
     nextcolor = 0
-    while curr_rev >= stop_rev:
+    while curr_rev is None or curr_rev >= stop_rev:
         # Compute revs and next_revs.
         if curr_rev not in revs:
             if branch:
                 ctx = repo.changectx(curr_rev)
                 if ctx.branch() != branch:
-                    curr_rev -= 1
+                    if curr_rev is None:
+                        curr_rev = len(repo.changelog)
+                    else:
+                        curr_rev -= 1
                     yield None
                     continue
+                    
             # New head.
+            if start_rev and follow and curr_rev != start_rev:
+                curr_rev -= 1
+                continue
             revs.append(curr_rev)
             rev_color[curr_rev] = curcolor = nextcolor
             nextcolor += 1
@@ -160,7 +174,11 @@ def revision_grapher(repo, start_rev=None, stop_rev=0, branch=None):
 
         yield (curr_rev, rev_index, curcolor, lines, parents)
         revs = next_revs
-        curr_rev -= 1
+        if curr_rev is None:
+            curr_rev = len(repo.changelog)
+        else:
+            curr_rev -= 1
+
 
 def filelog_grapher(repo, path):
     '''
@@ -175,7 +193,7 @@ def filelog_grapher(repo, path):
     heads = [repo.filectx(path, fileid=flog.rev(x)).rev() for x in flog.heads()]
     assert rev in heads
     heads.remove(rev)
-    
+
     revs = []
     rev_color = {}
     nextcolor = 0
@@ -220,14 +238,14 @@ def filelog_grapher(repo, path):
         yield (fctx.rev(), index, curcolor, lines, pcrevs,
                _paths.get(fctx.rev(), path))
         revs = next_revs
-        
+
         if revs:
             rev = max(revs)
         else:
             rev = -1
         if heads and rev <= heads[-1]:
             rev = heads.pop()
-            
+
 class GraphNode(object):
     """
     Simple class to encapsulate e hg node in the revision graph. Does
@@ -262,44 +280,56 @@ class Graph(object):
         self.nodesdict = {}
         self.max_cols = 0
 
-    def _build_nodes(self, nnodes):
-        """Internal method.
-        Build `nnodes` more nodes in our graph.
+    def build_nodes(self, nnodes=None, rev=None):
+        """
+        Build up to `nnodes` more nodes in our graph, or build as many
+        nodes required to reach `rev`.
+
+        If both rev and nnodes are set, build as many nodes as
+        required to reach rev plus nnodes more.
         """
         if self.grapher is None:
             return False
-
         stopped = False
         mcol = [self.max_cols]
-        for _ in xrange(nnodes):
-            try:
-                vnext = self.grapher.next()
-                if vnext is None:
-                    continue
-                nrev, xpos, color, lines, parents = vnext[:5]
-                if nrev >= self.maxlog:
-                    continue
-                gnode = GraphNode(nrev, xpos, color, lines, parents,
-                                  extra=vnext[5:])
-                if self.nodes:
-                    gnode.toplines = self.nodes[-1].bottomlines
-                self.nodes.append(gnode)
-                self.nodesdict[nrev] = gnode
-                mcol.append(gnode.cols)
-            except StopIteration:
-                self.grapher = None
-                stopped = True
-                break
+        for vnext in self.grapher:
+            if vnext is None:
+                continue
+            nrev, xpos, color, lines, parents = vnext[:5]
+            if nrev >= self.maxlog:
+                continue
+            gnode = GraphNode(nrev, xpos, color, lines, parents,
+                              extra=vnext[5:])
+            if self.nodes:
+                gnode.toplines = self.nodes[-1].bottomlines
+            self.nodes.append(gnode)
+            self.nodesdict[nrev] = gnode
+            mcol.append(gnode.cols)
+            if rev is not None and nrev <= rev:
+                rev = None # we reached rev, switching to nnode counter
+            if rev is None:
+                if nnodes is not None:
+                    nnodes -= 1
+                    if not nnodes:
+                        break
+                else:
+                    break
+        else:
+            self.grapher = None
+            stopped = True
 
         self.max_cols = max(mcol)
         return not stopped
 
+    def isfilled(self):
+        return self.grapher is None
+
     def fill(self, step=100):
         """
         Return a generator that fills the graph by bursts of `step`
         more nodes at each iteration.
         """
-        while self._build_nodes(step):
+        while self.build_nodes(step):
             yield len(self)
         yield len(self)
 
@@ -310,7 +340,7 @@ class Graph(object):
         if idx >= len(self.nodes):
             # build as many graph nodes as required to answer the
             # requested idx
-            self._build_nodes(idx)
+            self.build_nodes(idx)
         if idx > len(self):
             return self.nodes[-1]
         return self.nodes[idx]
@@ -320,16 +350,18 @@ class Graph(object):
         return len(self.nodes)
 
     def index(self, rev):
-        if rev < self.nodes[-1].rev:
-            self._build_nodes(self.nodes[-1].rev - rev)
+        if len(self) == 0: # graph is empty, let's build some nodes
+            self.build_nodes(10)
+        if rev is not None and rev < self.nodes[-1].rev:
+            self.build_nodes(self.nodes[-1].rev - rev)
         if rev in self.nodesdict:
             return self.nodes.index(self.nodesdict[rev])
         return -1
-        
+
     def fileflags(self, filename, rev):
         """
         Return a couple of flags ('=', '+', '-' or '?') depending on the nature
-        of the diff for filename between rev and its parents.        
+        of the diff for filename between rev and its parents.
         """
         ctx = self.repo.changectx(rev)
         flags = []
@@ -351,24 +383,33 @@ class Graph(object):
 
     def fileflag(self, filename, rev):
         """
-        Return a flag (see fileflags) between rev and its forst parent
+        Return a flag (see fileflags) between rev and its first parent
         """
         return self.fileflags(filename, rev)[0]
-    
+
     def filename(self, rev):
         return self.nodesdict[rev].extra[0]
-        
+
     def filedata(self, filename, rev, mode='diff'):
         """XXX written under dubious encoding assumptions
-        """
+        """        
         # XXX This really begins to be a dirty mess...
         data = ""
         flag = self.fileflag(filename, rev)
         ctx = self.repo.changectx(rev)
-
-        if flag not in ('-', '?'):
+        try:
             fctx = ctx.filectx(filename)
-            if fctx.node() is None:
+        except LookupError:
+            fctx = None # may happen for renamed files?
+
+        if isbfile(filename):
+            data = "[bfile]\n"
+            if fctx:
+                data = fctx.data()
+                data += "footprint: %s\n" % data
+            return "+", data
+        if flag not in ('-', '?'):
+            if fctx is None:# or fctx.node() is None:
                 return '', None
             if fctx.size() > self.maxfilesize:
                 data = "file too big"
@@ -409,7 +450,10 @@ class Graph(object):
         return flag, data
 
     def fileparent(self, filename, rev):
-        node = self.repo.changelog.node(rev)
+        if rev is not None:
+            node = self.repo.changelog.node(rev)            
+        else:
+            node = self.repo.changectx(rev).node()
         for parent in self.nodesdict[rev].parents:
             pnode = self.repo.changelog.node(parent)
             changes = self.repo.status(pnode, node)[:5]
