@@ -1,5 +1,5 @@
 # -*- coding: utf-8 -*-
-# Copyright (c) 2003-2009 LOGILAB S.A. (Paris, FRANCE).
+# Copyright (c) 2003-2010 LOGILAB S.A. (Paris, FRANCE).
 # http://www.logilab.fr/ -- mailto:contact@logilab.fr
 #
 # This program is free software; you can redistribute it and/or modify it under
@@ -29,6 +29,7 @@ SIGNAL = QtCore.SIGNAL
 Qt = QtCore.Qt
 
 from hgviewlib.config import HgConfig
+from hgviewlib.qt4 import should_rebuild
 
 class HgDialogMixin(object):
     """
@@ -44,9 +45,10 @@ class HgDialogMixin(object):
 
         _path = osp.dirname(__file__)
         uifile = osp.join(_path, self._uifile)
+        pyfile = uifile.replace(".ui", "_ui.py")
+        if should_rebuild(uifile, pyfile):
+            os.system('pyuic4 %s -o %s' % (uifile, pyfile))
         try:
-            ui_class, base_class = uic.loadUiType(uifile)
-        except IOError:
             modname = osp.splitext(osp.basename(uifile))[0] + "_ui"
             modname = "hgviewlib.qt4.%s" % modname
             mod = __import__(modname, fromlist=['*'])
@@ -56,7 +58,10 @@ class HgDialogMixin(object):
             elif 'Ui_MainWindow' in classnames:
                 ui_class = getattr(mod, 'Ui_MainWindow')
             else:
-                raise ValueError("Can't determine which main class to use in %s" % modname) 
+                raise ValueError("Can't determine which main class to use in %s" % modname)
+        except ImportError:
+            ui_class, base_class = uic.loadUiType(uifile)
+
         if ui_class not in self.__class__.__bases__:
             # hacking by adding the form class from ui file or pyuic4
             # generated module because we cannot use metaclass here,
@@ -64,7 +69,7 @@ class HgDialogMixin(object):
             self.__class__.__bases__ = self.__class__.__bases__ + (ui_class,)
         self.setupUi(self)
         self.load_ui()
-        self.disab_shortcuts = [self.esc_shortcut]
+        self.disab_shortcuts = []
         
     def load_ui(self):
         # we explicitely create a QShortcut so we can disable it
@@ -73,7 +78,7 @@ class HgDialogMixin(object):
         self.esc_shortcut = QtGui.QShortcut(self)
         self.esc_shortcut.setKey(Qt.Key_Escape)
         connect(self.esc_shortcut, SIGNAL('activated()'),
-                self.close)
+                self.maybeClose)
         self._quickbars = []
 
     def attachQuickBar(self, qbar):
@@ -94,6 +99,14 @@ class HgDialogMixin(object):
         for w in self._quickbars:
             if w is not tb:
                 w.hide()
+
+    def maybeClose(self):
+        for w in self._quickbars:
+            if w.isVisible():
+                w.cancel()
+                break
+        else:
+            self.close()
         
     def load_config(self):
         cfg = HgConfig(self.repo.ui)
