@@ -1,4 +1,4 @@
-# Copyright (c) 2009 LOGILAB S.A. (Paris, FRANCE).
+# Copyright (c) 2009-2010 LOGILAB S.A. (Paris, FRANCE).
 # http://www.logilab.fr/ -- mailto:contact@logilab.fr
 #
 # This program is free software; you can redistribute it and/or modify it under
@@ -19,16 +19,17 @@ Qt4 model for hg repo changelogs and filelogs
 import sys
 import mx.DateTime as dt
 import re
-import os.path as osp
+import os, os.path as osp
 
 from mercurial.node import nullrev
 from mercurial.node import hex, short as short_hex
 from mercurial.revlog import LookupError
+from mercurial import util, error
 
 from hgviewlib.hggraph import Graph, ismerge, diff as revdiff
 from hgviewlib.hggraph import revision_grapher, filelog_grapher
 from hgviewlib.config import HgConfig
-from hgviewlib.util import tounicode
+from hgviewlib.util import tounicode, isbfile, Curry
 from hgviewlib.qt4 import icon as geticon
 from hgviewlib.decorators import timeit
 
@@ -63,23 +64,44 @@ def cvrt_date(date):
     date, tzdelay = date
     return QtCore.QDateTime.fromTime_t(int(date)).toString(QtCore.Qt.LocaleDate)
 
+def gettags(model, ctx, gnode):
+    if ctx.rev() is None:
+        return ""
+    mqtags = ['qbase', 'qtip', 'qparent']
+    tags = ctx.tags()
+    if model.hide_mq_tags:
+        tags = [t for t in tags if t not in mqtags]
+    return ",".join(tags)
+
+def getlog(model, ctx, gnode):
+    if ctx.rev() is not None:
+        msg = tounicode(ctx.description())
+        if msg:
+            msg = msg.splitlines()[0]
+    else:
+        msg = "WORKING DIRECTORY (locally modified)"
+    return msg
+
+# XXX maybe it's time to make these methods of the model...
 # in following lambdas, ctx is a hg changectx
-_columnmap = {'ID': lambda ctx, gnode: ctx.rev(),
-              'Log': lambda ctx, gnode: tounicode(ctx.description()),
-              'Author': lambda ctx, gnode: tounicode(ctx.user()),
-              'Date': lambda ctx, gnode: cvrt_date(ctx.date()),
-              'Tags': lambda ctx, gnode: ",".join(ctx.tags()),
-              'Branch': lambda ctx, gnode: ctx.branch(),
-              'Filename': lambda ctx, gnode: gnode.extra[0],
+_columnmap = {'ID': lambda model, ctx, gnode: ctx.rev() is not None and str(ctx.rev()) or "",
+              'Log': getlog,
+              'Author': lambda model, ctx, gnode: tounicode(ctx.user()),
+              'Date': lambda model, ctx, gnode: cvrt_date(ctx.date()),
+              'Tags': gettags,
+              'Branch': lambda model, ctx, gnode: ctx.branch(),
+              'Filename': lambda model, ctx, gnode: gnode.extra[0],
               }
 
+_tooltips = {'ID': lambda model, ctx, gnode: ctx.rev() is not None and ctx.hex() or "Working Directory",
+             }
+
 def auth_width(model, repo):
     auths = model._aliases.values()
     if not auths:
         return None
     return sorted(auths, cmp=lambda x,y: cmp(len(x), len(y)))[-1]
 
-
 # in following lambdas, r is a hg repo
 _maxwidth = {'ID': lambda self, r: str(len(r.changelog)),
              'Date': lambda self, r: cvrt_date(r.changectx(0).date()),
@@ -119,73 +141,104 @@ class HgRepoListModel(QtCore.QAbstractTableModel):
     _stretchs = {'Log': 1, }
     _getcolumns = "getChangelogColumns"
 
-    def __init__(self, repo, branch='', parent=None):
+    def __init__(self, repo, branch='', fromhead=None, follow=False, parent=None):
         """
         repo is a hg repo instance
         """
         QtCore.QAbstractTableModel.__init__(self, parent)
         self._datacache = {}
-        self._required = None
-        self.wd_rev = None
-        self.gr_fill_timer = QtCore.QTimer()
-        connect(self.gr_fill_timer, SIGNAL('timeout()'),
-                self.fillGraph)
-        self.setRepo(repo, branch)
-
-    def setRepo(self, repo, branch=''):
+        self._hasmq = False
+        self.mqueues = []
+        self.wd_revs = []
+        self.graph = None
+        self._fill_timer = None
+        self.rowcount = 0
         self.repo = repo
-        self._datacache = {}
         self.load_config()
+        self.setRepo(repo, branch=branch, fromhead=fromhead, follow=follow)
 
-        self.wd_rev = self.repo.changectx(None).parents()[0].rev()
+    def setRepo(self, repo, branch='', fromhead=None, follow=False):
+        oldrepo = self.repo
+        self.repo = repo
+        if oldrepo.root != repo.root:
+            self.load_config()
+        self._datacache = {}
+        try:
+            wdctxs = self.repo.changectx(None).parents()
+        except error.Abort:
+            # might occur if reloading during a mq operation (or
+            # whatever operation playing with hg history)
+            return
+        self._hasmq = hasattr(self.repo, "mq")
+        if self._hasmq:
+            self.mqueues = self.repo.mq.series[:]
+        self.wd_revs = [ctx.rev() for ctx in wdctxs]
+        self.wd_status = [self.repo.status(ctx.node(), None)[:4] for ctx in wdctxs]
         self._user_colors = {}
         self._branch_colors = {}
-        grapher = revision_grapher(self.repo, branch=branch)
+        grapher = revision_grapher(self.repo, start_rev=fromhead,
+                                   follow=follow, branch=branch)
         self.graph = Graph(self.repo, grapher, self.max_file_size)
-        self.nmax = len(self.repo.changelog)
+        self.rowcount = 0
+        self.emit(SIGNAL('layoutChanged()'))
         self.heads = [self.repo.changectx(x).rev() for x in self.repo.heads()]
-        self._fill_iter = None
-        self._required = self.fill_step
-        self.gr_fill_timer.start()
+        self.ensureBuilt(row=self.fill_step)
+        QtCore.QTimer.singleShot(0, Curry(self.emit, SIGNAL('filled')))
+        self._fill_timer = self.startTimer(50)
 
-    def fillGraph(self):
-        step = self.fill_step
-        if self._fill_iter is None:
-            self.emit(SIGNAL('showMessage'), 'filling...')            
-            self._fill_iter = self.graph.fill(step=step)
-            self.emit(SIGNAL('layoutChanged()'))
-            QtGui.QApplication.processEvents()
-        try:
-            n = len(self.graph)
-            nm = min(n+step, self.nmax)
-            self.beginInsertRows(QtCore.QModelIndex(), n, nm)
-            nfilled = self._fill_iter.next()
-            if n == 0:
-                self.emit(SIGNAL('filled'))                
-            if self._required and len(self.graph) > self._required:
-                self._requiredss = None
-                self.gr_fill_timer.stop()
-                self.emit(SIGNAL('showMessage'), '')            
-        except StopIteration:
-            self.gr_fill_timer.stop()
-            self._fill_iter = None
-            self.emit(SIGNAL('showMessage'), '')            
-        finally:
+    def ensureBuilt(self, rev=None, row=None):
+        """
+        Make sure rev data is available (graph element created).
+
+        """
+        if self.graph.isfilled():
+            return
+        required = 0
+        buildrev = rev
+        n = len(self.graph)
+        if rev is not None:
+            if n and self.graph[-1].rev <= rev:
+                buildrev = None
+            else:
+                required = self.fill_step/2
+        elif row is not None and row > (n - self.fill_step / 2):
+            required = row - n + self.fill_step
+        if required or buildrev:
+            self.graph.build_nodes(nnodes=required, rev=buildrev)
+            self.updateRowCount()
+        elif row and row > self.rowcount:
+            # asked row was already built, but views where not aware of this
+            self.updateRowCount()
+        elif rev is not None and rev <= self.graph[self.rowcount].rev:
+            # asked rev was already built, but views where not aware of this
+            self.updateRowCount()
+
+    def timerEvent(self, event):
+        if event.timerId() == self._fill_timer:
+            self.emit(SIGNAL('showMessage'), 'filling (%s)'%(len(self.graph)))
+            if self.graph.isfilled():
+                self.killTimer(self._fill_timer)
+                self._fill_timer = None
+                self.emit(SIGNAL('showMessage'), '')
+            # we only fill the graph data strctures without telling
+            # views (until we atually did the full job), to keep
+            # maximal GUI reactivity
+            elif not self.graph.build_nodes(nnodes=self.fill_step):
+                self.killTimer(self._fill_timer)
+                self._fill_timer = None
+                self.updateRowCount()
+                self.emit(SIGNAL('showMessage'), '')
+
+    def updateRowCount(self):
+        currentlen = self.rowcount
+        newlen = len(self.graph)
+        if newlen > self.rowcount:
+            self.beginInsertRows(QtCore.QModelIndex(), currentlen, newlen-1)
+            self.rowcount = newlen
             self.endInsertRows()
-            self.emit(SIGNAL('layoutChanged()'))
-
-    def ensureBuilt(self, row=None, rev=None):
-        if self._fill_iter is not None and not self.gr_fill_timer.isActive():
-            if row is not None and (len(self.graph) - row) < self.fill_step:
-                self._required = len(self.graph) + self.fill_step
-            elif rev is not None and self.graph[-1].rev > rev:
-                self._required = len(self.graph) + self.fill_step + self.graph[-1].rev - rev
-            if self._required is not None:
-                self.gr_fill_timer.start()
-                self.emit(SIGNAL('showMessage'), 'filling...')            
 
     def rowCount(self, parent=None):
-        return len(self.graph)
+        return self.rowcount
 
     def columnCount(self, parent=None):
         return len(self._columns)
@@ -197,6 +250,7 @@ class HgRepoListModel(QtCore.QAbstractTableModel):
         self.rowheight = cfg.getRowHeight()
         self.fill_step = cfg.getFillingStep()
         self.max_file_size = cfg.getMaxFileSize()
+        self.hide_mq_tags = cfg.getMQHideTags()
         
         cols = getattr(cfg, self._getcolumns)()
         if cols is not None:
@@ -208,7 +262,6 @@ class HgRepoListModel(QtCore.QAbstractTableModel):
             elif 'Log' not in validcols or 'ID' not in validcols:
                 print "WARNING! 'Log' and 'ID' are mandatory. Check your configuration."
                 print "         reverting to default columns configuration"
-
             else:
                 self._columns = tuple(validcols)
 
@@ -255,16 +308,22 @@ class HgRepoListModel(QtCore.QAbstractTableModel):
         ctx = self.repo.changectx(gnode.rev)
         if role == QtCore.Qt.DisplayRole:
             if column == 'Author': #author
-                return QtCore.QVariant(self.user_name(_columnmap[column](ctx, gnode)))
+                return QtCore.QVariant(self.user_name(_columnmap[column](self, ctx, gnode)))
             elif column == 'Log':
-                msg = _columnmap[column](ctx, gnode)
-                if gnode.rev in self.heads:
-                    msg = "[%s] " % ctx.branch() + msg
+                msg = _columnmap[column](self, ctx, gnode)
                 return QtCore.QVariant(msg)
-            return QtCore.QVariant(_columnmap[column](ctx, gnode))
+            return QtCore.QVariant(_columnmap[column](self, ctx, gnode))
         elif role == QtCore.Qt.ToolTipRole:
-            msg = "<b>[%s]</b><br>\n" % ctx.branch()
-            msg += _columnmap[column](ctx, gnode)
+            msg = "<b>Branch:</b> %s<br>\n" % ctx.branch()
+            if gnode.rev in self.wd_revs:
+                msg += " <i>Working Directory position"
+                states = 'modified added removed deleted'.split()
+                status = self.wd_status[self.wd_revs.index(gnode.rev)]
+                status = [state for st, state in zip(status, states) if st]
+                if status:
+                    msg += ' (%s)' % (', '.join(status))
+                msg += "</i><br>\n"
+            msg += _tooltips.get(column, _columnmap[column])(self, ctx, gnode)
             return QtCore.QVariant(msg)
         elif role == QtCore.Qt.ForegroundRole:
             if column == 'Author': #author
@@ -319,8 +378,30 @@ class HgRepoListModel(QtCore.QAbstractTableModel):
                 pen = QtGui.QPen(pencolor)
                 pen.setWidth(penradius)
                 painter.setPen(pen)
-                if gnode.rev == self.wd_rev:
-                    painter.drawRect(dot_x, dot_y, radius, radius)
+                tags = set(ctx.tags())
+                icn = None
+
+                modified = False
+                atwd = False
+                if gnode.rev in self.wd_revs:
+                    atwd = True
+                    status = self.wd_status[self.wd_revs.index(gnode.rev)]
+                    if [True for st in status if st]:
+                        modified = True
+
+                if gnode.rev is None:
+                    # WD is displayed only if there are local
+                    # modifications, so let's use the modified icon
+                    icn = geticon('modified')
+                elif tags.intersection(self.mqueues):
+                    icn = geticon('mqpatch')
+                #elif modified:
+                #    icn = geticon('modified')
+                elif atwd:
+                    icn = geticon('clean')
+
+                if icn:
+                    icn.paint(painter, dot_x-5, dot_y-5, 17, 17)
                 else:
                     painter.drawEllipse(dot_x, dot_y, radius, radius)
                 painter.end()
@@ -365,34 +446,41 @@ class FileRevModel(HgRepoListModel):
     _stretchs = {'Log': 1, }
     _getcolumns = "getFilelogColumns"
 
-    def __init__(self, repo, filename, parent=None):
+    def __init__(self, repo, filename=None, parent=None):
         """
         data is a HgHLRepo instance
         """
         HgRepoListModel.__init__(self, repo, parent=parent)
         self.setFilename(filename)
 
-    def setRepo(self, repo, branch=''):
+    def setRepo(self, repo, branch='', fromhead=None, follow=False):
         self.repo = repo
         self._datacache = {}
         self.load_config()
 
     def setFilename(self, filename):
         self.filename = filename
-        self.nmax = len(self.repo.changelog)
+
         self._user_colors = {}
         self._branch_colors = {}
 
-        grapher = filelog_grapher(self.repo, self.filename)
-        self.graph = Graph(self.repo, grapher, self.max_file_size)
-        fl = self.repo.file(self.filename)
-        # we use fl.index here (instead of linkrev) cause
-        # linkrev API changed between 1.0 and 1.?. So this
-        # works with both versions.
-        self.heads = [fl.index[fl.rev(x)][4] for x in fl.heads()]
+        self.rowcount = 0
         self._datacache = {}
-        self._fill_iter = None
-        self.gr_fill_timer.start()
+
+        if self.filename:
+            grapher = filelog_grapher(self.repo, self.filename)
+            self.graph = Graph(self.repo, grapher, self.max_file_size)
+            fl = self.repo.file(self.filename)
+            # we use fl.index here (instead of linkrev) cause
+            # linkrev API changed between 1.0 and 1.?. So this
+            # works with both versions.
+            self.heads = [fl.index[fl.rev(x)][4] for x in fl.heads()]
+            self.ensureBuilt(row=self.fill_step/2)
+            QtCore.QTimer.singleShot(0, Curry(self.emit, SIGNAL('filled')))
+            self._fill_timer = self.startTimer(500)
+        else:
+            self.graph = None
+            self.heads = []
 
 
 replus = re.compile(r'^[+][^+].*', re.M)
@@ -412,6 +500,7 @@ class HgFileListModel(QtCore.QAbstractTableModel):
         self.load_config()
         self.current_ctx = None
         self._files = []
+        self._filesdict = {}
         self.diffwidth = 100
         self._fulllist = False
         self._fill_iter = None
@@ -435,8 +524,8 @@ class HgFileListModel(QtCore.QAbstractTableModel):
             self.diffwidth = w
             self._datacache = {}
             self.emit(SIGNAL('dataChanged(const QModelIndex &, const QModelIndex & )'),
-                      self.index(2, 0),
-                      self.index(2, self.rowCount()))
+                      self.index(1, 0),
+                      self.index(1, self.rowCount()))
 
     def __len__(self):
         return len(self._files)
@@ -473,19 +562,19 @@ class HgFileListModel(QtCore.QAbstractTableModel):
             if current_file_desc['fromside'] == 'right':
                 return self.current_ctx.parents()[1].rev()
             else:
-                return self.current_ctx.parents()[0].rev()                
+                return self.current_ctx.parents()[0].rev()
         return None
-    
+
     def indexFromFile(self, filename):
         if filename in self._filesdict:
             row = self._files.index(self._filesdict[filename])
             return self.index(row, 0)
         return QtCore.QModelIndex()
 
-    def _filterFile(self, filename):
+    def _filterFile(self, filename, ctxfiles):
         if self._fulllist:
             return True
-        return filename in self.current_ctx.files()
+        return filename in ctxfiles #self.current_ctx.files()
 
     def _buildDesc(self, parent, fromside):
         _files = []
@@ -493,32 +582,19 @@ class HgFileListModel(QtCore.QAbstractTableModel):
         ctxfiles = ctx.files()
         changes = self.repo.status(parent.node(), ctx.node())[:3]
         modified, added, removed = changes
-        for f in [x for x in added if self._filterFile(x)]:
-            desc = {'path':f, 'flag': '+', 'desc':f,
-                    'parent': parent, 'fromside': fromside,
-                    'infiles': f in ctxfiles}
-            m = ctx.filectx(f).renamed()
-            if m:
-                oldname, node = m
-                if oldname in removed:
-                    removed.remove(oldname)
-                    desc['renamedfrom'] = (oldname, node)
-                    desc['flag'] = '='
-                    desc['desc'] += '\n (was %s)' % oldname
-                else:
-                    desc['copiedfrom'] = (oldname, node)
-                    desc['flag'] = '='
-                    desc['desc'] += '\n (copy of %s)' % oldname
+        for lst, flag in ((added, '+'), (modified, '='), (removed, '-')):
+            for f in [x for x in lst if self._filterFile(x, ctxfiles)]:
+                _files.append({'path': f, 'flag': flag, 'desc': f,
+                               'parent': parent, 'fromside': fromside,
+                               'infiles': f in ctxfiles})
+                # renamed/copied files are handled by background
+                # filling process since it can be a bit long
+        for fdesc in _files:
+            bfile = isbfile(fdesc['path'])
+            fdesc['bfile'] = bfile
+            if bfile:
+                fdesc['desc'] = fdesc['desc'].replace('.hgbfiles'+os.sep, '')
 
-            _files.append(desc)
-        for f in [x for x in modified if self._filterFile(x)]:
-            _files.append({'path':f, 'flag': '=', 'desc':f,
-                           'parent': parent, 'fromside': fromside,
-                           'infiles': f in ctxfiles})
-        for f in [x for x in removed if self._filterFile(x)]:
-            _files.append({'path':f, 'flag': '-', 'desc':f,
-                           'parent': parent, 'fromside': fromside,
-                           'infiles': f in ctxfiles})
         return _files
 
     def loadFiles(self):
@@ -545,19 +621,20 @@ class HgFileListModel(QtCore.QAbstractTableModel):
         Method called to start the background process of computing
         file stats, which are to be displayed in the 'Stats' column
         """
-        if self._displaydiff:
-            self._fill_iter = self._fill()
-            self._fill_one_step()
+        self._fill_iter = self._fill()
+        self._fill_one_step()
 
     def _fill_one_step(self):
         if self._fill_iter is None:
             return
         try:
-            row = self._fill_iter.next()
-            idx = self.index(row, 2)
-            self.emit(SIGNAL('dataChanged(const QModelIndex&, const QModelIndex&)'),
-                      idx, idx)
-            QtCore.QTimer.singleShot(0, lambda self=self: self._fill_one_step())
+            nextfill = self._fill_iter.next()
+            if nextfill is not None:
+                row, col = nextfill
+                idx = self.index(row, col)
+                self.emit(SIGNAL('dataChanged(const QModelIndex &, const QModelIndex &)'),
+                          idx, idx)
+            QtCore.QTimer.singleShot(10, lambda self=self: self._fill_one_step())
 
         except StopIteration:
             self._fill_iter = None
@@ -566,7 +643,7 @@ class HgFileListModel(QtCore.QAbstractTableModel):
         # the generator used to fill file stats as a background process
         for row, desc in enumerate(self._files):
             filename = desc['path']
-            if desc['flag'] == '=':
+            if desc['flag'] == '=' and self._displaydiff:
                 diff = revdiff(self.repo, self.current_ctx, desc['parent'],
                                files=[filename])
                 try:
@@ -578,9 +655,26 @@ class HgFileListModel(QtCore.QAbstractTableModel):
                 if tot == 0:
                     tot = max(add + rem, 1)
                 desc['stats'] = (tot, add, rem)
-            yield row
+                yield row, 1
+
+            if desc['flag'] == '+':
+                m = self.current_ctx.filectx(filename).renamed()
+                if m:
+                    removed = self.repo.status(desc['parent'].node(),
+                                               self.current_ctx.node())[2]
+                    oldname, node = m
+                    if oldname in removed:
+                        # removed.remove(oldname) XXX
+                        desc['renamedfrom'] = (oldname, node)
+                        desc['flag'] = '='
+                        desc['desc'] += '\n (was %s)' % oldname
+                    else:
+                        desc['copiedfrom'] = (oldname, node)
+                        desc['flag'] = '='
+                        desc['desc'] += '\n (copy of %s)' % oldname
+                    yield row, 0
+            yield None
 
-    @datacached
     def data(self, index, role):
         if not index.isValid() or index.row()>len(self) or not self.current_ctx:
             return nullvariant
