@@ -1,4 +1,4 @@
-# Copyright (c) 2009 LOGILAB S.A. (Paris, FRANCE).
+# Copyright (c) 2009-2010 LOGILAB S.A. (Paris, FRANCE).
 # http://www.logilab.fr/ -- mailto:contact@logilab.fr
 #
 # This program is free software; you can redistribute it and/or modify it under
@@ -31,8 +31,10 @@ SIGNAL = QtCore.SIGNAL
 nullvariant = QtCore.QVariant()
 
 from hgviewlib.decorators import timeit
+from hgviewlib.config import HgConfig
+from hgviewlib.util import format_desc, xml_escape
 from hgviewlib.qt4 import icon as geticon
-from hgviewlib.qt4.hgfileviewer import ManifestViewer
+from hgviewlib.qt4.hgmanifestdialog import ManifestViewer
 from hgviewlib.qt4.quickbar import QuickBar
 
 class GotoQuickBar(QuickBar):
@@ -66,6 +68,10 @@ class GotoQuickBar(QuickBar):
             self.entry.setFocus()
             self.entry.selectAll()
 
+    def __del__(self):
+        # prevent a warning in the console:
+        # QObject::startTimer: QTimer can only be used with threads started with QThread
+        self.entry.setCompleter(None)
 
 class HgRepoView(QtGui.QTableView):
     """
@@ -81,7 +87,7 @@ class HgRepoView(QtGui.QTableView):
         self.setSelectionMode(QtGui.QAbstractItemView.SingleSelection)
         self.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)
         self.setAlternatingRowColors(True)
-                
+
         self.createActions()
         self.createToolbars()
         connect(self,
@@ -92,7 +98,16 @@ class HgRepoView(QtGui.QTableView):
         connect(self.horizontalHeader(),
                 SIGNAL('sectionResized(int, int, int)'),
                 self.disableAutoResize)
-        
+
+    def mousePressEvent(self, event):
+        index = self.indexAt(event.pos())
+        if not index.isValid():
+            return
+        if event.button() == Qt.MidButton:
+            self.gotoAncestor(index)
+            return
+        QtGui.QTableView.mousePressEvent(self, event)
+
     def createToolbars(self):
         self.goto_toolbar = GotoQuickBar(self)
         connect(self.goto_toolbar, SIGNAL('goto'),
@@ -108,7 +123,12 @@ class HgRepoView(QtGui.QTableView):
     def createActions(self):
         self._actions = {}
         for name, desc, icon, tip, key, cb in self._action_defs():
-            act = QtGui.QAction(desc, self)
+            self._actions[name] = QtGui.QAction(desc, self)
+        QtCore.QTimer.singleShot(0, self.configureActions)
+
+    def configureActions(self):
+        for name, desc, icon, tip, key, cb in self._action_defs():
+            act = self._actions[name]
             if icon:
                 act.setIcon(geticon(icon))
             if tip:
@@ -117,7 +137,6 @@ class HgRepoView(QtGui.QTableView):
                 act.setShortcut(key)
             if cb:
                 connect(act, SIGNAL('triggered()'), cb)
-            self._actions[name] = act
             self.addAction(act)
 
     def showAtRev(self):
@@ -130,7 +149,6 @@ class HgRepoView(QtGui.QTableView):
                 menu.addAction(self._actions[act])
             else:
                 menu.addSeparator()
-        menu.addAction(self.goto_toolbar._actions['open'])
         menu.exec_(event.globalPos())
 
     def init_variables(self):
@@ -194,24 +212,27 @@ class HgRepoView(QtGui.QTableView):
                 w = model._stretchs[model._columns[c]] / tot_stretch
                 self.setColumnWidth(c, col1_width * w)
 
-    def revisionActivated(self, index):
+    def revFromindex(self, index):
         if not index.isValid():
             return
         model = self.model()
         if model and model.graph:
             row = index.row()
             gnode = model.graph[row]
-            self.emit(SIGNAL('revisionActivated'), gnode.rev)
+            return gnode.rev
+
+    def revisionActivated(self, index):
+        rev = self.revFromindex(index)
+        if rev is not None:
+            self.emit(SIGNAL('revisionActivated'), rev)
 
     def revisionSelected(self, index, index_from):
         """
         Callback called when a revision is selected in the revisions table
         """
-        model = self.model()
-        if model and model.graph:
-            row = index.row()
-            gnode = model.graph[row]
-            rev = gnode.rev
+        rev = self.revFromindex(index)
+        if True:#rev is not None:
+            model = self.model()
             if self.current_rev is not None and self.current_rev == rev:
                 return
             if not self._in_history:
@@ -225,6 +246,17 @@ class HgRepoView(QtGui.QTableView):
             self.emit(SIGNAL('revisionSelected'), rev)
             self.set_navigation_button_state()
 
+    def gotoAncestor(self, index):
+        rev = self.revFromindex(index)
+        if rev is not None and self.current_rev is not None:
+            repo = self.model().repo
+            ctx = repo[self.current_rev]
+            ctx2 = repo[rev]
+            ancestor = ctx.ancestor(ctx2)
+            self.emit(SIGNAL('showMessage'),
+                      "Goto ancestor of %s and %s"%(ctx.rev(), ctx2.rev()), 2000)
+            self.goto(ancestor.rev())
+
     def set_navigation_button_state(self):
         if len(self._rev_history) > 0:
             back = self._rev_pos > 0
@@ -265,8 +297,7 @@ class HgRepoView(QtGui.QTableView):
         else:
             idx = self.model().indexFromRev(rev)
             if idx is not None:
-                if self.sender() == self.goto_toolbar:
-                    self.sender().setVisible(False)
+                self.goto_toolbar.setVisible(False)
                 self.setCurrentIndex(idx)
 
     def nextRev(self):
@@ -299,28 +330,82 @@ class RevDisplay(QtGui.QTextBrowser):
             self.refreshDisplay()
             # TODO: emit a signal to recompute the diff
             self.emit(SIGNAL('parentRevisionSelected'), self.diffrev)
-        else:            
+        else:
             self.emit(SIGNAL('revisionSelected'), int(rev))
 
     def setDiffRevision(self, rev):
         if rev != self.diffrev:
             self.diffrev = rev
             self.refreshDisplay()
-    
+
     def displayRevision(self, ctx):
         self.ctx = ctx
         self.diffrev = ctx.parents()[0].rev()
+        if hasattr(self.ctx._repo, "mq"):
+            self.mqseries = self.ctx._repo.mq.series[:]
+            self.mqunapplied = [x[1] for x in self.ctx._repo.mq.unapplied(self.ctx._repo)]
+            mqpatch = set(self.ctx.tags()).intersection(self.mqseries)            
+            if mqpatch:
+                self.mqpatch = mqpatch.pop()
+            else:
+                self.mqpatch = None
+        else:
+            self.mqseries = []
+            self.mqunapplied = []
+            self.mqpatch = None
+            
         self.refreshDisplay()
+
+    def selectNone(self):
+        cursor = self.textCursor()
+        cursor.clearSelection()
+        cursor.setPosition(0)
+        self.setTextCursor(cursor)
+        self.setExtraSelections([])
+        
+    def searchString(self, text):
+        self.selectNone()
+        if text in unicode(self.toPlainText()):
+            clist = []
+            while self.find(text):
+                eselect = self.ExtraSelection()
+                eselect.cursor = self.textCursor()
+                eselect.format.setBackground(QtGui.QColor('#ffffbb'))
+                clist.append(eselect)
+            self.selectNone()
+            self.setExtraSelections(clist)
+            def finditer(self, text):
+                if text:
+                    while True:
+                        if self.find(text):
+                            yield self.ctx.rev(), None                
+                        else:
+                            break
+            return finditer(self, text)
         
     def refreshDisplay(self):
         ctx = self.ctx
         rev = ctx.rev()
         buf = "<table width=100%>\n"
+        if self.mqpatch:
+            buf += '<tr bgcolor=%s>' % HgConfig(ctx._repo.ui).getMQFGColor()
+            buf += '<td colspan=3 width=100%><b>Patch queue:</b>&nbsp;'
+            for p in self.mqseries:
+                if p in self.mqunapplied:
+                    p = "<i>%s</i>" % p
+                elif p == self.mqpatch:
+                    p = "<b>%s</b>" % p
+                buf += '&nbsp;%s&nbsp;' % (p)
+            buf += '</td></tr>\n'
+
         buf += '<tr>'
-        buf += '<td><b>Revision:</b>&nbsp;'\
-               '<span class="rev_number">%d</span>:'\
-               '<span class="rev_hash">%s</span></td>'\
-               '\n' % (ctx.rev(), short_hex(ctx.node()))
+        if rev is None:
+            buf += "<td><b>Working Directory</b></td>\n"
+        else:
+            buf += '<td><b>Revision:</b>&nbsp;'\
+                   '<span class="rev_number">%d</span>:'\
+                   '<span class="rev_hash">%s</span></td>'\
+                   '\n' % (ctx.rev(), short_hex(ctx.node()))
 
         buf += '<td><b>Author:</b>&nbsp;'\
                '%s</td>'\
@@ -333,9 +418,7 @@ class RevDisplay(QtGui.QTextBrowser):
         for p in parents:
             if p.rev() > -1:
                 short = short_hex(p.node())
-                desc = unicode(p.description(), 'utf-8', 'replace')
-                if len(desc) > self.descwidth:
-                    desc = desc[:self.descwidth] + '...'
+                desc = format_desc(p.description(), self.descwidth)
                 p_rev = p.rev()
                 p_fmt = '<span class="rev_number">%s</span>:'\
                         '<a href="%s" class="rev_hash">%s</a>'
@@ -350,9 +433,7 @@ class RevDisplay(QtGui.QTextBrowser):
         if len(parents) == 2:
             p = parents[0].ancestor(parents[1])
             short = short_hex(p.node())
-            desc = unicode(p.description(), 'utf-8', 'replace')
-            if len(desc) > self.descwidth:
-                desc = desc[:self.descwidth] + '...'
+            desc = format_desc(p.description(), self.descwidth)
             p_rev = p.rev()
             p_fmt = '<span class="rev_number">%s</span>:'\
                     '<a href="%s" class="rev_hash">%s</a>'
@@ -364,13 +445,11 @@ class RevDisplay(QtGui.QTextBrowser):
                    '<td colspan=5>%s&nbsp;'\
                    '<span class="short_desc"><i>%s</i></span></td></tr>'\
                    '\n' % (p_rev, desc)
-            
+
         for p in ctx.children():
             if p.rev() > -1:
                 short = short_hex(p.node())
-                desc = unicode(p.description(), 'utf-8', 'replace')
-                if len(desc) > self.descwidth:
-                    desc = desc[:self.descwidth] + '...'
+                desc = format_desc(p.description(), self.descwidth)
                 buf += '<tr><td class="label"><b>Child:</b></td>'\
                        '<td colspan=5><span class="rev_number">%d</span>:'\
                        '<a href="%s" class="rev_hash">%s</a>&nbsp;'\
@@ -378,7 +457,8 @@ class RevDisplay(QtGui.QTextBrowser):
                        '\n' % (p.rev(), p.rev(), short, desc)
 
         buf += "</table>\n"
-        desc = unicode(ctx.description().replace('\n', '<br/>\n'), 'utf-8', 'replace')
+        desc = xml_escape(unicode(ctx.description(), 'utf-8', 'replace'))
+        desc = desc.replace('\n', '<br/>\n')
         buf += '<div class="diff_desc"><p>%s</p></div>\n' % desc
         self.setHtml(buf)
 
