@@ -2,7 +2,7 @@
 #!/usr/bin/env python
 # main.py - qt4-based hg rev log browser
 #
-# Copyright (C) 2007-2009 Logilab. All rights reserved.
+# Copyright (C) 2007-2010 Logilab. All rights reserved.
 #
 # This software may be used and distributed according to the terms
 # of the GNU General Public License, incorporated herein by reference.
@@ -18,12 +18,14 @@ from mercurial import ui, hg
 from mercurial import util
 
 from hgviewlib.util import tounicode, has_closed_branch_support
+from hgviewlib.util import rootpath, find_repository
 from hgviewlib.hggraph import diff as revdiff
 from hgviewlib.decorators import timeit
 
 from hgviewlib.qt4 import icon as geticon
 from hgviewlib.qt4.hgrepomodel import HgRepoListModel, HgFileListModel
-from hgviewlib.qt4.hgfileviewer import ManifestViewer
+from hgviewlib.qt4.hgfiledialog import FileViewer, FileDiffViewer
+from hgviewlib.qt4.hgmanifestdialog import ManifestViewer
 from hgviewlib.qt4.hgdialogmixin import HgDialogMixin
 from hgviewlib.qt4.quickbar import FindInGraphlogQuickBar
 from hgviewlib.qt4.helpviewer import HelpViewer
@@ -42,7 +44,7 @@ SIGNAL = QtCore.SIGNAL
 class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin):
     """hg repository viewer/browser application"""
     _uifile = 'hgqv.ui'
-    def __init__(self, repo):
+    def __init__(self, repo, fromhead=None):
         self.repo = repo
         self._closed_branch_supp = has_closed_branch_support(self.repo)
 
@@ -50,7 +52,6 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin):
         self._reload_rev = None
         self._reload_file = None
 
-
         QtGui.QMainWindow.__init__(self)
         HgDialogMixin.__init__(self)
 
@@ -74,8 +75,9 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin):
         connect(self.textview_status, SIGNAL('fileDisplayed'),
                 self.file_displayed)
         self.setupBranchCombo()
-        self.setupModels()
-
+        self.setupModels(fromhead)
+        if fromhead:
+            self.startrev_entry.setText(str(fromhead))
         self.setupRevisionTable()
 
         self._repodate = self._getrepomtime()
@@ -87,6 +89,7 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin):
             if mtime > self._repodate:
                 self.statusBar().showMessage("Repository has been modified, "
                                              "reloading...", 2000)
+
                 self.reload()
 
     def setupBranchCombo(self, *args):
@@ -97,27 +100,25 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin):
                 openbr.extend(self.repo.branchheads(branch, closed=False))
             clbranches = [br for br, node in allbranches if node not in openbr]
             branches = [br for br, node in allbranches if node in openbr]
-            self.branch_checkBox_action.setVisible(len(clbranches)>0)
-            if self.branch_checkBox.isChecked():
+            if self.branch_checkBox_action.isChecked():
                 branches = branches + clbranches
         else:
-            self.branch_checkBox_action.setVisible(False)
             branches = [br for br, node in allbranches] # open branches
 
         if len(branches) == 1:
-            self.branch_label_action.setVisible(False)
-            self.branch_comboBox_action.setVisible(False)
+            self.branch_label_action.setEnabled(False)
+            self.branch_comboBox_action.setEnabled(False)
         else:
             self.branchesmodel = QtGui.QStringListModel([''] + branches)
             self.branch_comboBox.setModel(self.branchesmodel)
-            self.branch_label_action.setVisible(True)
             self.branch_label_action.setEnabled(True)
-            self.branch_comboBox_action.setVisible(True)
             self.branch_comboBox_action.setEnabled(True)
 
     def createToolbars(self):
+        # find quickbar
         self.find_toolbar = FindInGraphlogQuickBar(self)
         self.find_toolbar.attachFileView(self.textview_status)
+        self.find_toolbar.attachHeaderView(self.textview_header)
         connect(self.find_toolbar, SIGNAL('revisionSelected'),
                 self.tableView_revisions.goto)
         connect(self.find_toolbar, SIGNAL('fileSelected'),
@@ -128,22 +129,63 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin):
 
         self.attachQuickBar(self.find_toolbar)
 
+        # navigation toolbar
         self.toolBar_edit.addAction(self.tableView_revisions._actions['back'])
         self.toolBar_edit.addAction(self.tableView_revisions._actions['forward'])
 
-        self.branch_label = QtGui.QLabel("Branch")
+        findaction = self.find_toolbar.toggleViewAction()
+        findaction.setIcon(geticon('find'))
+        self.toolBar_edit.addAction(findaction)
+
+        # tree filters toolbar
+        self.branch_label = QtGui.QToolButton()
+        self.branch_label.setText("Branch")
+        self.branch_label.setStatusTip("Display graph the named branch only")
+        self.branch_label.setPopupMode(QtGui.QToolButton.InstantPopup)
+        self.branch_menu = QtGui.QMenu()
+        cbranch_action = self.branch_menu.addAction("Display closed branches")
+        cbranch_action.setCheckable(True)
+        self.branch_checkBox_action = cbranch_action
+        self.branch_label.setMenu(self.branch_menu)
         self.branch_comboBox = QtGui.QComboBox()
-        self.branch_checkBox = QtGui.QCheckBox("Display closed branches")
         connect(self.branch_comboBox, SIGNAL('activated(const QString &)'),
                 self.refreshRevisionTable)
-        connect(self.branch_checkBox, SIGNAL('toggled(bool)'),
+        connect(cbranch_action, SIGNAL('toggled(bool)'),
                 self.setupBranchCombo)
 
-        self.branch_checkBox_action = self.toolBar_treefilters.addWidget(self.branch_checkBox)
-        self.toolBar_treefilters.addSeparator()
+        self.toolBar_treefilters.layout().setSpacing(3)
+
         self.branch_label_action = self.toolBar_treefilters.addWidget(self.branch_label)
         self.branch_comboBox_action = self.toolBar_treefilters.addWidget(self.branch_comboBox)
+        self.toolBar_treefilters.addSeparator()
+
+        self.startrev_label = QtGui.QToolButton()
+        self.startrev_label.setText("Start rev.")
+        self.startrev_label.setStatusTip("Display graph from this revision")
+        self.startrev_label.setPopupMode(QtGui.QToolButton.InstantPopup)
+        self.startrev_entry = QtGui.QLineEdit()
+        self.startrev_entry.setStatusTip("Display graph from this revision")
+        self.startrev_menu = QtGui.QMenu()
+        follow_action = self.startrev_menu.addAction("Follow mode")
+        follow_action.setCheckable(True)
+        follow_action.setStatusTip("Follow changeset history from start revision")
+        self.startrev_follow_action = follow_action
+        self.startrev_label.setMenu(self.startrev_menu)
+
+        self.revscompl_model = QtGui.QStringListModel(['tip'])
+        self.revcompleter = QtGui.QCompleter(self.revscompl_model, self)
+        self.startrev_entry.setCompleter(self.revcompleter)
+
+        connect(self.startrev_entry, SIGNAL('editingFinished()'),
+                self.refreshRevisionTable)
+        connect(self.startrev_follow_action, SIGNAL('toggled(bool)'),
+                self.refreshRevisionTable)
+        self.startrev_label_action = self.toolBar_treefilters.addWidget(self.startrev_label)
+        self.startrev_entry_action = self.toolBar_treefilters.addWidget(self.startrev_entry)
+
+        # diff mode toolbar
         self.toolBar_diff.addAction(self.actionDiffMode)
+        self.toolBar_diff.addAction(self.actionAnnMode)
         self.toolBar_diff.addAction(self.actionNextDiff)
         self.toolBar_diff.addAction(self.actionPrevDiff)
 
@@ -163,8 +205,13 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin):
         connect(self.actionDiffMode, SIGNAL('toggled(bool)'),
                 self.setMode)
 
+        self.actionAnnMode = QtGui.QAction('Annotate', self)
+        self.actionAnnMode.setCheckable(True)
+        connect(self.actionAnnMode, SIGNAL('toggled(bool)'),
+                self.textview_status.setAnnotate)
+
         self.actionHelp.setShortcut(Qt.Key_F1)
-        self.actionHelp.setIcon(geticon('help'))        
+        self.actionHelp.setIcon(geticon('help'))
         connect(self.actionHelp, SIGNAL('triggered()'),
                 self.on_help)
         
@@ -207,7 +254,6 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin):
         self.disab_shortcuts.append(self.actionNextRev)
         self.disab_shortcuts.append(self.actionPrevRev)
 
-
         # navigate in file viewer
         self.actionNextLine = QtGui.QAction('Next line', self)
         self.actionNextLine.setShortcut(Qt.SHIFT + Qt.Key_Down)
@@ -233,9 +279,15 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin):
         # Activate file (file diff navigator)
         self.actionActivateFile = QtGui.QAction('Activate file', self)
         self.actionActivateFile.setShortcuts([Qt.Key_Return, Qt.Key_Enter])
+        def enterkeypressed():
+            w = QtGui.QApplication.focusWidget()
+            if not isinstance(w, QtGui.QLineEdit):
+                self.tableView_filelist.fileActivated(self.tableView_filelist.currentIndex(),)
+            else:
+                w.emit(SIGNAL('editingFinished()'))
         connect(self.actionActivateFile, SIGNAL('triggered()'),
-                lambda self=self:
-                self.tableView_filelist.fileActivated(self.tableView_filelist.currentIndex(),))
+                enterkeypressed)
+
         self.actionActivateFileAlt = QtGui.QAction('Activate alt. file', self)
         self.actionActivateFileAlt.setShortcuts([Qt.ALT+Qt.Key_Return, Qt.ALT+Qt.Key_Enter])
         connect(self.actionActivateFileAlt, SIGNAL('triggered()'),
@@ -252,9 +304,35 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin):
         self.disab_shortcuts.append(self.actionActivateFile)
         self.disab_shortcuts.append(self.actionActivateRev)
 
+        self.actionStartAtRev = QtGui.QAction('Start at rev.', self)
+        self.actionStartAtRev.setShortcuts([Qt.Key_Backspace,])
+        connect(self.actionStartAtRev, SIGNAL('triggered()'),
+                self.startAtCurrentRev)
+        self.addAction(self.actionStartAtRev)
+
+        self.actionClearStartAtRev = QtGui.QAction('Clear start at rev.', self)
+        self.actionClearStartAtRev.setShortcuts([Qt.SHIFT + Qt.Key_Backspace,])
+        connect(self.actionClearStartAtRev, SIGNAL('triggered()'),
+                self.clearStartAtRev)
+        self.addAction(self.actionClearStartAtRev)
+
+    def startAtCurrentRev(self):
+        crev = self.tableView_revisions.current_rev
+        if crev:
+            self.startrev_entry.setText(str(crev))
+            # XXX workaround: see refreshRevisionTable method 
+            self.refreshRevisionTable(sender=self)
+
+    def clearStartAtRev(self):
+        self.startrev_entry.setText("")
+        self._reload_rev = self.tableView_revisions.current_rev
+        self._reload_file = self.tableView_filelist.currentFile()
+        # XXX workaround: see refreshRevisionTable method 
+        self.refreshRevisionTable(sender=self)
 
     def setMode(self, mode):
         self.textview_status.setMode(mode)
+        self.actionAnnMode.setEnabled(not mode)
         self.actionNextDiff.setEnabled(not mode)
         self.actionPrevDiff.setEnabled(not mode)
 
@@ -272,8 +350,8 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin):
         cfg = HgDialogMixin.load_config(self)
         self.hidefinddelay = cfg.getHideFindDelay()
 
-    def create_models(self):
-        self.repomodel = HgRepoListModel(self.repo)
+    def create_models(self, fromhead=None):
+        self.repomodel = HgRepoListModel(self.repo, fromhead=fromhead)
         connect(self.repomodel, SIGNAL('filled'),
                 self.on_filled)
         connect(self.repomodel, SIGNAL('showMessage'),
@@ -282,8 +360,8 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin):
 
         self.filelistmodel = HgFileListModel(self.repo)
 
-    def setupModels(self):
-        self.create_models()
+    def setupModels(self, fromhead=None):
+        self.create_models(fromhead)
         self.tableView_revisions.setModel(self.repomodel)
         self.tableView_filelist.setModel(self.filelistmodel)
         self.textview_status.setModel(self.repomodel)
@@ -303,7 +381,10 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin):
         connect(self.textview_header, SIGNAL('revisionSelected'), view.goto)
         connect(self.textview_header, SIGNAL('parentRevisionSelected'), self.textview_status.displayDiff)
         self.attachQuickBar(view.goto_toolbar)
-
+        gotoaction = view.goto_toolbar.toggleViewAction()
+        gotoaction.setIcon(geticon('goto'))
+        self.toolBar_edit.addAction(gotoaction)
+        
     def _setup_table(self, table):
         table.setTabKeyNavigation(False)
         table.verticalHeader().setDefaultSectionSize(self.rowheight)
@@ -363,6 +444,14 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin):
             if len(self.filelistmodel):
                 self.tableView_filelist.selectRow(0)
 
+    def goto(self, rev):
+        if len(self.tableView_revisions.model().graph):
+            self.tableView_revisions.goto(rev)
+        else:
+            # store rev to show once it's available (when graph
+            # filling is still running)
+            self._reload_rev = rev
+            
     def _getrepomtime(self):
         """Return the last modification time for the repo"""
         watchedfiles = [(self.repo.root, ".hg", "store", "00changelog.i"),
@@ -383,15 +472,30 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin):
         self._repodate = self._getrepomtime()
         self.setupBranchCombo()
         self.setupModels()
-        self.refreshRevisionTable()
+        # XXX workaround: see refreshRevisionTable method 
+        self.refreshRevisionTable(sender=self)
 
     #@timeit
-    def refreshRevisionTable(self, branch=None):
+    def refreshRevisionTable(self, *args, **kw):
         """Starts the process of filling the HgModel"""
-        if branch is None:
-            branch = self.branch_comboBox.currentText()
+        branch = self.branch_comboBox.currentText()
         branch = str(branch)
-        self.repomodel.setRepo(self.repo, branch=branch)
+        startrev = str(self.startrev_entry.text()).strip()
+        if not startrev:
+            startrev = None
+        # XXX workaround: self.sender() may provoque a core dump if
+        # this method is called directly (not via a connected signal);
+        # the 'sender' keyword is a way to discrimimne that the method
+        # has been called directly (thus caller MUST set this kw arg)
+        sender = kw.get('sender') or self.sender()
+        if sender is self.startrev_follow_action and startrev is None:
+            return
+        startrev = self.repo.changectx(startrev).rev()
+        follow = self.startrev_follow_action.isChecked()
+        self.revscompl_model.setStringList(self.repo.tags().keys())
+
+        self.repomodel.setRepo(self.repo, branch=branch, fromhead=startrev,
+                               follow=follow)
 
     def on_about(self, *args):
         """ Display about dialog """
@@ -414,18 +518,6 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin):
         w.raise_()
         w.activateWindow()
         
-def find_repository(path):
-    """returns <path>'s mercurial repository
-
-    None if <path> is not under hg control
-    """
-    path = os.path.abspath(path)
-    while not os.path.isdir(os.path.join(path, ".hg")):
-        oldpath = path
-        path = os.path.dirname(path)
-        if path == oldpath:
-            return None
-    return path
 
 def main():
     from optparse import OptionParser
@@ -451,17 +543,17 @@ def main():
                       action="store_true",
                       help='(with filename) start in navigation mode')
 
-    opt, args = parser.parse_args()
+    opts, args = parser.parse_args()
 
-    if opt.navigate and len(args) != 1:
+    if opts.navigate and len(args) != 1:
         parser.error("You must provide a filename to start in navigate mode")
 
     if len(args) > 1:
         parser.error("Provide at most one file name")
 
     dir_ = None
-    if opt.repo:
-        dir_ = opt.repo
+    if opts.repo:
+        dir_ = opts.repo
     else:
         dir_ = os.getcwd()
     dir_ = find_repository(dir_)
@@ -477,15 +569,21 @@ def main():
     signal.signal(signal.SIGINT, signal.SIG_DFL)
 
     app = QtGui.QApplication(sys.argv)
-
+    from hgviewlib.qt4 import setup_font_substitutions
+    setup_font_substitutions()
+    
     if len(args) == 1:
+        filename = rootpath(repo, opts.rev, args[0])
+        if filename is None:
+            parser.error("%s is not a tracked file" % args[0])
+        
         # should be a filename of a file managed in the repo
-        if opt.get('navigate'):
-            mainwindow = FileViewer(repo, args[0])
+        if opts.navigate:
+            mainwindow = FileViewer(repo, filename)
         else:
-            mainwindow = FileDiffViewer(repo, args[0])
+            mainwindow = FileDiffViewer(repo, filename)
     else:
-        rev = opt.rev
+        rev = opts.rev
         if rev is not None:
             try:
                 repo.changectx(rev)
